package ru.ispras.telecomtesk.tlsv12.HandshakeMessages;

import java.math.BigInteger;
import java.security.Key;
import java.security.KeyFactory;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.Security;
import java.security.cert.X509Certificate;
import java.security.spec.EncodedKeySpec;
import java.util.Random;
import javax.crypto.Cipher;
import javax.crypto.KeyAgreement;
import javax.crypto.interfaces.DHKey;
import javax.crypto.interfaces.DHPublicKey;
import javax.crypto.spec.DHParameterSpec;
import javax.crypto.spec.DHPublicKeySpec;
import javax.crypto.spec.IvParameterSpec;
import javax.crypto.spec.SecretKeySpec;

import org.bouncycastle.jce.provider.BouncyCastleProvider;

import ru.ispras.telecomtesk.tlsv12.DebugInfo;
import ru.ispras.telecomtesk.tlsv12.TLSAlgorithms.*;
import ru.ispras.telecomtesk.tlsv12.TLSState.TLSState;
import ru.ispras.telecomtesk.tlsv12.HandshakeMessages.KeyExchange.ServerDHParams;
import ru.ispras.telecomtesk.tlsv12.RecordProtocol.ProtocolVersion;
import ru.ispras.telecomtesk.tlsv12.RecordProtocol.GeneratePairKeys;

public class ClientKeyExchange extends HandshakeMessage
{
	public KeyExchangeAlgorithm key_alg;
	public ProtocolVersion client_version;
	public byte[] random_bytes; // 46 bytes
	public byte[] EncryptedPreMasterSecret;
	public boolean implicitPublicValueEncoding = true;
	public byte[] DH_Yc;
	public int error = 0;
	private boolean set_signature = false;
	
	public ClientKeyExchange(KeyExchangeAlgorithm key_alg)
	{ 
		msg_type = 16; 
		this.key_alg = key_alg;
		if(key_alg == KeyExchangeAlgorithm.RSA) {
			client_version = new ProtocolVersion();
			random_bytes = new byte[46];
			Random rndObj = new Random(System.currentTimeMillis());
			rndObj.nextBytes(random_bytes);
		}
		else
			random_bytes = null;
		EncryptedPreMasterSecret = null;
		DH_Yc = null;
	}

	public ClientKeyExchange(KeyExchangeAlgorithm key_alg, byte[] data, boolean isRSAencryptedPreMasterSecret)
	{ 
		msg_type = 16; 
		this.key_alg = key_alg;
		random_bytes = null;
		EncryptedPreMasterSecret = null;
		DH_Yc = null;
		
		switch (key_alg) {
		case RSA:
			if(isRSAencryptedPreMasterSecret) {
				EncryptedPreMasterSecret = data;
				set_signature = true;
			}
			else {
				client_version = new ProtocolVersion();
				random_bytes = new byte[46];
				Random rndObj = new Random(System.currentTimeMillis());
				rndObj.nextBytes(random_bytes);
			}
			break;
		case DH_DSS:
		case DH_RSA:
		case DH_ANON:
		case DHE_DSS:
		case DHE_RSA:
			if(data != null) {
				this.DH_Yc = data;
				implicitPublicValueEncoding = false;
			}
		}
	}

	public ClientKeyExchange(KeyExchangeAlgorithm key_alg,
							 ProtocolVersion ver,
			                 byte[] rnd)
	{ 
		msg_type = 16; 
		this.key_alg = key_alg;
		client_version = ver;
		EncryptedPreMasterSecret = null;
		DH_Yc = null;

		if(rnd.length != 46) 
			System.out.println("ClientKeyExchange: byte[].length != 46");
		random_bytes = rnd.length == 46 ? rnd : new byte[46];
	}

	public void setEncryptedPreMasterSecret(byte[] pms) {
		set_signature = true;
		EncryptedPreMasterSecret = pms;
	}
	
	//  RFC requires to use PKCS1 padding. For testing purpose also used PKCS5 padding
	public void setEncryptedPreMasterSecret(Key skey, boolean isPKCS1Padding) throws Exception {
		// calculate EncryptedPreMasterSecret; if failed, set error = -1
		byte[] preMasterSecret = new byte[48];
		preMasterSecret[0] = client_version.major;
		preMasterSecret[1] = client_version.minor;
		System.arraycopy(random_bytes, 0, preMasterSecret, 2, 46);
		Cipher cobj;
		
		if(skey != null) {
			Security.addProvider(new BouncyCastleProvider());

			if(isPKCS1Padding == true)
				// RSA/ECB/PKCS1Padding
				cobj = Cipher.getInstance("RSA/NONE/PKCS1Padding","BC");
			else
				cobj = Cipher.getInstance("RSA/NONE/PKCS5Padding","BC");
			
			cobj.init(Cipher.ENCRYPT_MODE, skey);
			EncryptedPreMasterSecret = cobj.doFinal(preMasterSecret); 

			set_signature = true;
		}
	}
	
	//  RFC requires to use PKCS1 padding. For testing purpose also used PKCS5 padding
	public void setEncryptedPreMasterSecret(byte[] preMasterSecret, Key skey, boolean isPKCS1Padding) throws Exception {
		// calculate EncryptedPreMasterSecret; if failed, set error = -1
		Cipher cobj;
		
		DebugInfo.printDump(4,preMasterSecret,"ClientKeyExchange::setEncryptedPreMasterSecret: external preMasterSecret");

		Security.addProvider(new BouncyCastleProvider());

		if(isPKCS1Padding == true)
			// RSA/ECB/PKCS1Padding
			cobj = Cipher.getInstance("RSA/NONE/PKCS1Padding","BC");
		else
			cobj = Cipher.getInstance("RSA/NONE/PKCS5Padding","BC");
			
		cobj.init(Cipher.ENCRYPT_MODE, skey);
		EncryptedPreMasterSecret = cobj.doFinal(preMasterSecret); 

		set_signature = true;
	}
	
	public void setEncryptedPreMasterSecret(X509Certificate cert, boolean isPKCS1Padding) throws Exception {
		// calculate EncryptedPreMasterSecret; if failed, set error = -1
		byte[] preMasterSecret = new byte[48];
		preMasterSecret[0] = client_version.major;
		preMasterSecret[1] = client_version.minor;
		System.arraycopy(random_bytes, 0, preMasterSecret, 2, 46);
		Cipher cobj;
		
		Security.addProvider(new BouncyCastleProvider());
			
		if(isPKCS1Padding == true)
			// RSA/ECB/PKCS1Padding
			cobj = Cipher.getInstance("RSA/NONE/PKCS1Padding","BC");
		else
			cobj = Cipher.getInstance("RSA/NONE/PKCS5Padding","BC");
			
		cobj.init(Cipher.ENCRYPT_MODE, cert);
		EncryptedPreMasterSecret = cobj.doFinal(preMasterSecret); 

		set_signature = true;
	}
	
	public void decryptEncryptedPreMasterSecret(Key skey) throws Exception {
		// calculate EncryptedPreMasterSecret; if failed, set error = -1
		byte[] preMasterSecret = null;
		random_bytes = null;
		Cipher cobj;

		DebugInfo.printMessage(4, "ClientKeyExchange::decryptEncryptedPreMasterSecret start");
		
		if(skey != null) {
			Security.addProvider(new BouncyCastleProvider());
			cobj = Cipher.getInstance("RSA/NONE/PKCS1Padding","BC");
			cobj.init(Cipher.DECRYPT_MODE, skey);
			preMasterSecret = cobj.doFinal(EncryptedPreMasterSecret);
			DebugInfo.printMessage(1,"===decryptEncryptedPreMasterSecret:");
			DebugInfo.printDump(1, preMasterSecret, null);
 
			if(preMasterSecret != null) {
				if(preMasterSecret.length < 2) {
					error = -1;
					DebugInfo.printMessage(1, "===Error!: ClientKeyExchange::decryptEncryptedPreMasterSecret: decrypted data length failed: " + preMasterSecret.length);
				}
				else {
					DebugInfo.printMessage(4, "decryptEncryptedPreMasterSecret: decrypted data length: " + preMasterSecret.length);
					client_version = new ProtocolVersion(preMasterSecret[0], preMasterSecret[1]);
					random_bytes = new byte[preMasterSecret.length - 2];
					if(random_bytes.length > 0)
						System.arraycopy(preMasterSecret, 2, random_bytes, 0, random_bytes.length);
				}
			}
			else {
				error = -1;
				DebugInfo.printMessage(1, "===Error!: ClientKeyExchange::decryptEncryptedPreMasterSecret: decrypt failed: result null");
			}
		}
		else {
			error = -1;
			DebugInfo.printMessage(1, "===Error!: ClientKeyExchange::decryptEncryptedPreMasterSecret: key for decrypt isn't defined");
		}

	}
	
	public void setDiffieHellmanValue(byte[] dhv) {
		set_signature = true;
		implicitPublicValueEncoding = false;
		DH_Yc = dhv;
	}

	public byte[] setDiffieHellmanValue(ServerDHParams serverDHparams) {
		byte[] DH_Secret = null;
		
		switch (key_alg) {
			case DH_ANON:
			case DHE_DSS:
			case DHE_RSA:
				if(serverDHparams.isDHempty() == false) {
			        try {
						BigInteger p = new BigInteger(1,serverDHparams.dh_p);
						BigInteger g = new BigInteger(1,serverDHparams.dh_g);
						BigInteger yserv = new BigInteger(1,serverDHparams.dh_Ys);
			            DHPublicKeySpec dhKeySpec = new DHPublicKeySpec(yserv, p, g);
			            KeyFactory serverKeyFactory = KeyFactory.getInstance("DiffieHellman");
			            DHPublicKey serverPublicKey = (DHPublicKey) serverKeyFactory.generatePublic(dhKeySpec);
					    DHParameterSpec param = serverPublicKey.getParams();
			            KeyPairGenerator keyGen = KeyPairGenerator.getInstance("DiffieHellman");
			            keyGen.initialize(param);
			            KeyPair kp = keyGen.genKeyPair();
			            DHPublicKey clientPublicKey = (DHPublicKey)(kp.getPublic()); 

						DH_Yc = clientPublicKey.getY().toByteArray();
						
						KeyAgreement clAgree = KeyAgreement.getInstance("DiffieHellman");
						clAgree.init(kp.getPrivate());
						clAgree.doPhase(serverPublicKey, true);
			            DH_Secret = clAgree.generateSecret();
			            
			            DebugInfo.printMessage(6,"ClientKeyExchange::setDiffieHellmanValue: Diffie-Hellman values");
			            DebugInfo.printDump(6, serverDHparams.dh_g, "serverDHparams.dh_g");
			            DebugInfo.printDump(6, serverDHparams.dh_p, "serverDHparams.dh_p");
			            DebugInfo.printDump(6, serverDHparams.dh_Ys, "serverDHparams.dh_Ys");
			            DebugInfo.printDump(6, DH_Yc, "DH_Yc");
			            DebugInfo.printDump(5, DH_Secret, "DH_Secret");
			            // trim first zero bytes
			            int i;
			            for(i=0;i<DH_Secret.length;i++)
			            	if(DH_Secret[i] != 0) break;
			            if(i>0)
			            	System.arraycopy(DH_Secret, i, DH_Secret, 0, DH_Secret.length - i);
			        } catch (Exception e) {
						DebugInfo.printMessage(1,"===Error! GeneratePairKeys::generateKeysDH: " + e);
			        }

			        set_signature = true;
					implicitPublicValueEncoding = false;
				}
		}
		return DH_Secret;
	}

	public int getSize() {
		switch (key_alg) {
			case RSA:
				if((set_signature == false)||(EncryptedPreMasterSecret == null)) 
					error = 10;
				if(error == 0)
					return (6 + (short)EncryptedPreMasterSecret.length);
				else
					return 0;
			case DH_DSS:
			case DH_RSA:
			case DH_ANON:
			case DHE_DSS:
			case DHE_RSA:
				if(implicitPublicValueEncoding == false) 
					return (6 + (short)DH_Yc.length);
				else
					return 4;
		}
		return 0;
	}
	public String toString() {
		String res = "ClientKeyExchange(16): ";
		if(msg_type != 16)
			res += "!!!failed HandshakeMessage type " + msg_type + " ";
		
		if(EncryptedPreMasterSecret != null)
			res += "EncryptedPreMasterSecret";
		else if(DH_Yc != null){
			if(implicitPublicValueEncoding)
				res += "implicit DH_Yc";
			else
				res += "explicit DH_Yc";
		}
		else
			res += "null key materials";
			
		return res;
	}
	public int toByteArray() {
		int i=0,len, res = 0;
		if(isLengthSet == false) setLength();
		// create byte array
		byte[] ba = new byte[getSize()];
		net_msg = ba;
		if(ba.length == 0) return res;

		// write msg_type, full length and version
		byte[] ba_len = hton(length);
		ba[0] = msg_type;
		System.arraycopy(ba_len, 1, ba, 1, 3);
		i=4;
		
		switch (key_alg) {
		case RSA:
			len = (short)EncryptedPreMasterSecret.length;
			ba[i++] = (byte)(len >>> 8);
			ba[i++] = (byte)len;
			System.arraycopy(EncryptedPreMasterSecret, 0, ba, i, len);
			break;
		case DH_ANON:
		case DHE_DSS:
		case DHE_RSA:
			len = (short)DH_Yc.length;
			ba[i++] = (byte)(len >>> 8);
			ba[i++] = (byte)len;
			System.arraycopy(DH_Yc, 0, ba, i, len);
			break;
		case DH_DSS:
		case DH_RSA:
			break;

		}
		return res;
	}
}
