package ru.ispras.telecomtesk.tlsv12.HandshakeMessages;

//need add crypto methods
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.KeyPair;
import java.security.Security;
import java.security.Signature;
import java.security.cert.Certificate;

import javax.crypto.interfaces.DHPublicKey;
import javax.crypto.spec.DHParameterSpec;

import org.bouncycastle.jce.provider.BouncyCastleProvider;

import ru.ispras.telecomtesk.tlsv12.DebugInfo;
import ru.ispras.telecomtesk.tlsv12.HandshakeMessages.KeyExchange.*;
import ru.ispras.telecomtesk.tlsv12.RecordProtocol.DigitallySigned;
import ru.ispras.telecomtesk.tlsv12.TLSAlgorithms.*;
import ru.ispras.telecomtesk.tlsv12.TLSState.TLSSecurityParameters;


public class ServerKeyExchange extends HandshakeMessage
{
	public KeyExchangeAlgorithm key_alg;
	public ServerDHParams params;
	public byte[] client_random = new byte[32];
	public byte[] server_random = new byte[32];
	public DigitallySigned ds;
	public int error = 0;
	private boolean set_signature = false;
	
	public ServerKeyExchange(KeyExchangeAlgorithm key_alg,
							 byte[] dh_p,
							 byte[] dh_g,
							 byte[] dh_Ys,
			                 byte[] clnt_rnd,
			                 byte[] srv_rnd,
			                 SignatureAndHashAlgorithm sah) 
	{ 
		msg_type = 12; 
		this.key_alg = key_alg;
		params = new ServerDHParams(dh_p, dh_g, dh_Ys);
		int len = clnt_rnd.length <= 32 ? clnt_rnd.length : 32;
		System.arraycopy(clnt_rnd, 0, client_random, 0, len);
		len = srv_rnd.length <= 32 ? srv_rnd.length : 32;
		System.arraycopy(srv_rnd, 0, server_random, 0, len);
		ds = new DigitallySigned(sah);
	}

	public ServerKeyExchange(KeyExchangeAlgorithm key_alg,
			 				 byte[] dh_p,
			 				 byte[] dh_g,
			 				 byte[] dh_Ys,
			 				 DigitallySigned ds) 
	{ 
		msg_type = 12; 
		this.key_alg = key_alg;
		params = new ServerDHParams(dh_p, dh_g, dh_Ys);
		this.ds = ds;
		set_signature = true;
	}

	public ServerKeyExchange(TLSSecurityParameters tlsparams, PrivateKey prvKey, KeyPair kp)
	{ 
		msg_type = 12; 
		DebugInfo.printMessage(5,"ServertKeyExchange: create full message");
		if((tlsparams != null) && (prvKey != null) && (kp != null)) {
			try {
				DebugInfo.printMessage(5,"ServertKeyExchange: create ServerDHParams");
				DHPublicKey pubKey = (DHPublicKey)(kp.getPublic());
				DHParameterSpec dhparam = pubKey.getParams();
				params = new ServerDHParams(dhparam.getP().toByteArray(), dhparam.getG().toByteArray(), pubKey.getY().toByteArray());

				DebugInfo.printMessage(5,"ServertKeyExchange: fill randoms");
				this.key_alg = tlsparams.ke_alg;
				int len = tlsparams.client_random.length <= 32 ? tlsparams.client_random.length : 32;
				System.arraycopy(tlsparams.client_random, 0, client_random, 0, len);
				len = tlsparams.server_random.length <= 32 ? tlsparams.server_random.length : 32;
				System.arraycopy(tlsparams.server_random, 0, server_random, 0, len);

				String algPrvKeyName = prvKey.getAlgorithm();
				DebugInfo.printMessage(5,"ServertKeyExchange: calculate signature, algorithm " + algPrvKeyName);
				if(algPrvKeyName.equalsIgnoreCase("DSA"))
					this.ds = new DigitallySigned(HashAlgorithm.SHA1, SignatureAlgorithm.DSA);
				else if(algPrvKeyName.equalsIgnoreCase("RSA"))
					this.ds = new DigitallySigned(HashAlgorithm.SHA1, SignatureAlgorithm.RSA);
				else
					throw new IllegalArgumentException("Unknown PrivateKey Algorithm: " + algPrvKeyName);
				
				setSignature(prvKey);
			} catch (Exception err) {
				DebugInfo.printMessage(1,"===Error! ServertKeyExchange: create message failed" + err);
				error = 1;
			}
		}
		else
			error = 1;
	}

	public void setSignature(byte[] sig) {
		set_signature = true;
		ds.signature = sig;
	}
	public void setSignature(PrivateKey skey) {
		// calculate ds.signature; if failed, set error = -1
		Signature sig;
		error = -1;
		DebugInfo.printMessage(5,"ServertKeyExchange::setSignature with key");
		if(params.isDHempty() == false && ds != null && skey != null) {
			String alg = HashAlgorithm.valueOf(ds.alg.hash) + "with" + SignatureAlgorithm.valueOf(ds.alg.signature);
            DebugInfo.printMessage(3, "ServertKeyExchange::setSignature: algorithm " + alg);
			byte[] dhparam = params.toByteArray();
			int datalen = client_random.length + server_random.length + dhparam.length;
			byte[] data = new byte[datalen];
			int pos = 0;
			System.arraycopy(client_random, 0, data, pos, client_random.length);
			pos += client_random.length;
			System.arraycopy(server_random, 0, data, pos, server_random.length);
			pos += server_random.length;
			System.arraycopy(dhparam, 0, data, pos, dhparam.length);
            DebugInfo.printDump(6, data, "signed data");
			try {
				Security.addProvider(new BouncyCastleProvider());
				sig = Signature.getInstance(alg,"BC");
				sig.initSign(skey);
				sig.update(data);
				this.ds.setSignature(sig.sign());
				set_signature = true;
				error = 0;
			} catch (Exception err) {
				DebugInfo.printMessage(1,"===Error! ServertKeyExchange::setSignature: " + err);
			}
		}
		DebugInfo.printMessage(3,"ServertKeyExchange::setSignature: " + error);
	}

	public byte[] getServerDh_p() {
		if(params != null && params.dh_p != null)
			return params.dh_p;
		else
			return null;
	}

	public byte[] getServerDh_g() {
		if(params != null && params.dh_g != null)
			return params.dh_g;
		else
			return null;
	}

	public byte[] getServerDh_Ys() {
		if(params != null && params.dh_Ys != null)
			return params.dh_Ys;
		else
			return null;
	}

	public boolean validateDS(TLSSecurityParameters tlsparams, PublicKey skey) {
		Signature sig;
		boolean res = false;
		DebugInfo.printMessage(5,"ServertKeyExchange::validateDS");
		if(params.isDHempty() == false && ds != null && ds.signature != null && skey != null && tlsparams != null) {
			String alg = HashAlgorithm.valueOf(ds.alg.hash) + "with" + SignatureAlgorithm.valueOf(ds.alg.signature);
            DebugInfo.printMessage(3, "ServertKeyExchange::validateDS: algorithm " + alg);
			byte[] dhparam = params.toByteArray();
			int datalen = tlsparams.client_random.length + tlsparams.server_random.length + dhparam.length;
			byte[] data = new byte[datalen];
			int pos = 0;
			System.arraycopy(tlsparams.client_random, 0, data, pos, tlsparams.client_random.length);
			pos += tlsparams.client_random.length;
			System.arraycopy(tlsparams.server_random, 0, data, pos, tlsparams.server_random.length);
			pos += tlsparams.server_random.length;
			System.arraycopy(dhparam, 0, data, pos, dhparam.length);
            DebugInfo.printDump(6, data, "signed data");
			try {
				Security.addProvider(new BouncyCastleProvider());
				sig = Signature.getInstance(alg,"BC");
				sig.initVerify(skey);
				sig.update(data);
				res = sig.verify(ds.signature);
			} catch (Exception err) {
				DebugInfo.printMessage(1,"===Error! ServertKeyExchange::validateDS: " + err);
			}
		}
		DebugInfo.printMessage(3,"ServertKeyExchange::validateDS: " + res);
		return res;
	}
	
	public int getSize() {
		switch (key_alg) {
			case RSA:
			case DH_DSS:
			case DH_RSA:
				return 0;
			case DH_ANON:
				return (4 + params.getSize());
			case DHE_DSS:
			case DHE_RSA:
				if(set_signature == true) 
					return (8 + params.getSize() + (short)ds.signature.length);
				else
					return 0;
		}
		return 0;
	}

	public String toString() {
		String res;

		res = "HandshakeMessage: type ServerKeyExchange, length " + length;
		if(net_msg != null)
			res += ", net_msg: " + net_msg.length;
		if(ds != null) {
			res += "\n      " + ds.alg.toString();
		}

		return res;
	}

	public int toByteArray() {
		if(isLengthSet == false) setLength();
		// create byte array
		byte[] ba = new byte[getSize()];
		int res = 0;

		switch (key_alg) {
		case RSA:
		case DH_DSS:
		case DH_RSA:
			break;
		case DH_ANON:
			int i = 0, j;
			// write msg_type, full length and version
			byte[] ba_len = hton(length);
			ba[0] = msg_type;
			System.arraycopy(ba_len, 1, ba, 1, 3);
			i=4;
			System.arraycopy(params.toByteArray(), 0, ba, i, params.getSize());
			i += params.getSize();
			break;
		case DHE_DSS:
		case DHE_RSA:
			if(set_signature == true) {
				i = 0;
				// write msg_type, full length and version
				ba_len = hton(length);
				ba[0] = msg_type;
				System.arraycopy(ba_len, 1, ba, 1, 3);
				i=4;
				System.arraycopy(params.toByteArray(), 0, ba, i, params.getSize());
				i += params.getSize();
				ba[i++] = ds.alg.hash;
				ba[i++] = ds.alg.signature;
				ba[i++] = (byte)(ds.signature.length >>> 8);
				ba[i++] = (byte)ds.signature.length;
				System.arraycopy(ds.signature, 0, ba, i, (short)ds.signature.length);
				break;
			}
			else {
				ba = new byte[0]; res = 1; break; }
		}
		net_msg = ba;
		return res;
	}
}
