package ru.ispras.telecomtesk.tlsv12.TLSSocket;

import java.io.*;
import java.net.*;
import java.util.*;
import javax.net.ssl.SSLSocket;
import javax.net.ssl.SSLSocketFactory;

import ru.ispras.telecomtesk.tlsv12.DebugInfo;

public class TLSServerChildSocket implements Runnable 
{
    private Socket skt = null;
    private BufferedOutputStream out = null;
    private DataInputStream in = null;

    protected long responseTimeout = 10000; // in milliseconds
    
    private byte[][] receivedData = new byte[10][];
    private int lastind;
    private int lastpos;
    private int curind;
    private int curpos;
    private int TLSRESPONSE_ARRAY_LENGTH = 65565;

    private String threadName;
    public Thread t = null;
    public volatile boolean runSocketFlag = true;
    public volatile boolean waitSocketFlag = false;

    
    public TLSServerChildSocket(Socket skt, String threadName)  throws IOException {
    	this.skt = skt;
    	this.threadName = threadName;
    	initReceivedData();
    	receivedData[0] = new byte[TLSRESPONSE_ARRAY_LENGTH];
    	DebugInfo.printMessage(1,"New socket created");
    	t = new Thread(this,threadName);
    	t.start();
    }

    public void initReceivedData() {
    	lastind = 0; lastpos = 0; curind = 0; curpos = 0; 
    }

	public String getIP() {
		return "" + skt.getRemoteSocketAddress();
	}

	public int getPort() {
		return skt.getPort();
	}

	public long getTimeout() {
		return responseTimeout;
	}

	public boolean getRunSocketFlag() {
		return runSocketFlag;
	}

	public void setRunSocketFlag(boolean flag) {
		this.runSocketFlag = flag;
	}

	public boolean getWaitSocketFlag() {
		return waitSocketFlag;
	}

	public void setWaitSocketFlag(boolean flag) {
		this.waitSocketFlag = flag;
	}

	public void setTimeout(long timeout) throws SocketException {
		responseTimeout = timeout;
		if(skt != null)
			skt.setSoTimeout((int)responseTimeout);
	}

	public void disconnect() {
    		runSocketFlag = false;
		if (skt != null){
			try {
				skt.close();
			}
			catch (IOException ignored) {
			}
		}
	}
	
    public void sendMessage(byte[] message) throws IOException
    {
    	DebugInfo.printMessage(5, "sendMessage start time: " + System.currentTimeMillis());
    	DebugInfo.printDump(5, message, "TLSSocket::sendMessage");

    	try {
    		out.write(message,0,message.length);
    		out.flush();
    	}
    	catch(IOException ioerr) {
			DebugInfo.printMessage(1,"===Error! TLSSocket::sendMessage: " + ioerr);
    		throw ioerr;
    	}
    }
    
    
    public void shiftPosition(int n) {
		DebugInfo.printMessage(5,"shiftPosition");
    	lastpos += n;
    	
		if(lastpos > TLSRESPONSE_ARRAY_LENGTH)
			throw new ArrayIndexOutOfBoundsException( "shiftPosition: bad curpos value");
		else if(lastpos == TLSRESPONSE_ARRAY_LENGTH) {
	    	lastpos = 0;
			lastind++;
			if(receivedData[lastind] == null)
				receivedData[lastind] = new byte[TLSRESPONSE_ARRAY_LENGTH];
		}
    }


    public void run() 
    {
    	long curtime, starttime;
    	int numbytes = -1;

    	try {
			out = new BufferedOutputStream(new DataOutputStream(skt.getOutputStream()));
			in = new DataInputStream(skt.getInputStream());

			starttime = System.currentTimeMillis();
			DebugInfo.printMessage(5, "TLSServerChildSocket: waitForResponse starttime " + starttime + " responseTimeout " + responseTimeout);
    	
			while (runSocketFlag) {
        		if(in.available() <= 0) continue;

        		synchronized (this) {
            		numbytes = in.read(receivedData[lastind],lastpos,receivedData[lastind].length - lastpos);
            		DebugInfo.printMessage(4, "TLSServerChildSocket: received " + numbytes + " bytes, time: " + System.currentTimeMillis());

            		if(numbytes >0) shiftPosition(numbytes);
    		
            		curtime = System.currentTimeMillis();
            		DebugInfo.printMessage(5, "TLSServerChildSocket: get first data time " + curtime);
        		}
			}
        
			out.close();
			in.close();
    	} catch (IOException err) {
   			DebugInfo.printMessage(1, err.toString());
   			runSocketFlag = false;;
    	}

    	DebugInfo.printMessage(5, "TLSServerChildSocket: socket stoped: curtime " + System.currentTimeMillis());
    	DebugInfo.printMessage(6, "TLSServerChildSocket: response " + lastind + " " + lastpos);
    }
    
   
    public TLSResponse getResponse() {
    	DebugInfo.printMessage(5, "TLSServerChildSocket: getResponse");
        TLSResponse socketResponse = new TLSResponse();
        waitSocketFlag = true;

        synchronized (this) {
        	DebugInfo.printMessage(5, "TLSServerChildSocket: getResponse synchronized block");
        	byte[][] tmpData = socketResponse.message;
        	socketResponse.message = receivedData;
        	receivedData = tmpData;
        	socketResponse.curind = this.curind;
        	socketResponse.curpos = this.curpos;
        	socketResponse.lastind = this.lastind;
        	socketResponse.lastpos = this.lastpos;
        	initReceivedData();
        	DebugInfo.printMessage(6, "TLSServerChildSocket: getResponse synchronized exit");
        }
        
        waitSocketFlag = false;

    	DebugInfo.printMessage(5, "TLSServerChildSocket: return");
    	return socketResponse;
    }

    public void tlswait(){
		DebugInfo.printMessage(5, "TLSServerChildSocket: tlswait");
    	while(waitSocketFlag) {
    		try {
    			DebugInfo.printMessage(5, "TLSServerChildSocket: wait");
    			wait();
    			DebugInfo.printMessage(5, "TLSServerChildSocket: continue");
    		} catch (InterruptedException err) {
    			DebugInfo.printMessage(1, "InterruptedException " +  + System.currentTimeMillis() + " " + err);
    		}
    	}
    }
    
    public void doPause(double iTimeInSeconds){
        long t0, t1;
        t0=System.currentTimeMillis( );
        t1=System.currentTimeMillis( )+Math.round(iTimeInSeconds*1000);

        do {
	        t0=System.currentTimeMillis( );
	    } while (t0 < t1);
    }
    
    public void close() {
    	DebugInfo.printMessage(4, "TLSServerChildSocket: socket " + t.getName() + " is stopping");
    	runSocketFlag = false;
    	try {
    		skt.close();
        	DebugInfo.printMessage(4, "TLSServerChildSocket: closed");
   	} catch (Exception ignored) {
    	}
    }

    public void finilize() {
    	runSocketFlag = false;
    	try {
    		out.close();
    		in.close();
    		skt.close();
    	} catch (Exception ignored) {
    	}
    }
   
}