package ru.ispras.telecomtesk.tlsv12.TLSState;

import java.net.Socket;

import ru.ispras.telecomtesk.tlsv12.DebugInfo;
import ru.ispras.telecomtesk.tlsv12.TLSAlgorithms.TLSCipherSuite;

public class TLSConnection 
{
	public Socket skt;
	public TLSConnectionSelector selector;
	TLSState write_current_state;
	TLSState read_current_state;
	TLSState write_pending_state;
	TLSState read_pending_state;
	public boolean isServer;
	public enum StateDirection {READ_STATE, WRITE_STATE};

	public TLSConnection(TLSConnectionSelector sel) {
		selector = sel;
		isServer = false;
		write_current_state = null;
		read_current_state = null;
		write_pending_state = null;
		read_pending_state = null;
	}
	public TLSConnection(TLSConnectionSelector sel, int init) {
		selector = sel;
		isServer = false;
		init();
	}
	public TLSConnection(TLSConnectionSelector sel, boolean isServer) {
		selector = sel;
		this.isServer = isServer;
		write_current_state = null;
		read_current_state = null;
		write_pending_state = null;
		read_pending_state = null;
	}
	public TLSConnection(TLSConnectionSelector sel, boolean isServer, int init) {
		selector = sel;
		this.isServer = isServer;
		init();
	}
	public TLSConnection(TLSConnectionSelector sel, TLSState wcs, TLSState wps, TLSState rcs, TLSState rps) {
		selector = sel;
		write_current_state = wcs;
		read_current_state = rcs;
		write_pending_state = wps;
		read_pending_state = rps;
		if((wcs != null) && (wcs.params != null))
			isServer = wcs.params.isServer;
		else
			isServer = false;
	}
	public void init() {
		write_current_state = new TLSState(this.isServer);
		read_current_state = new TLSState(this.isServer);
		write_pending_state = new TLSState(this.isServer);
		read_pending_state = new TLSState(this.isServer);
	}
	public void setSocketParameters(Socket skt) {
		if(skt != null) {
			this.skt = skt;
			selector = new TLSConnectionSelector(
					skt.getLocalAddress(),
					skt.getInetAddress(),
					skt.getLocalPort(),
					skt.getPort()
					);
		}
	}
	public TLSState getWriteCurrentState() {
		return write_current_state;
	}
	public TLSState getWritePendingState() {
		return write_pending_state;
	}
	public TLSState getReadCurrentState() {
		return read_current_state;
	}
	public TLSState getReadPendingState() {
		return read_pending_state;
	}
	public void setWriteCurrentState(TLSState st) {
		if(st != null)
			write_current_state = st;
		else
					DebugInfo.printMessage(1,"===Warning! TLSConnection::setWriteCurrentState NULL pointer. Reject set new state.");
	}
	public void changeState(StateDirection mode) {
		switch(mode) {
		case WRITE_STATE:
			write_current_state = write_pending_state;
			write_pending_state = new TLSState(this.isServer);
			break;
		case READ_STATE:
			read_current_state = read_pending_state;
			read_pending_state = new TLSState(this.isServer);
			break;
		}
	}
	public void initPendingState(StateDirection mode) {
		switch(mode) {
		case WRITE_STATE:
			write_pending_state = new TLSState(this.isServer);
			break;
		case READ_STATE:
			read_pending_state = new TLSState(this.isServer);
			break;
		}
	}
	public void initAllPendingStates() {
		write_pending_state = new TLSState(this.isServer);
		read_pending_state = new TLSState(this.isServer);
	}
	
	// Pending States functions
	public void setRandom(byte[] rnd, boolean isClientRandom) {
		write_pending_state.getParams().setRandom(rnd, isClientRandom);
		read_pending_state.getParams().setRandom(rnd, isClientRandom);
	}
	// set ID for !!!pending_state
	public void setSessionID(byte[] id) {
		write_pending_state.setSessionID(id);
		read_pending_state.setSessionID(id);
	}
	// get ID of !!!current_state
	public byte[] getSessionID() {
		return write_current_state.getSessionID();
	}
	public void setCompressionAlgorithm(byte alg) {
		write_pending_state.getParams().compression_alg = alg;
		read_pending_state.getParams().compression_alg = alg;
	}
	public void setAlgorithmParams(TLSCipherSuite cs) {
		write_pending_state.getParams().setAlgorithmParams(cs);
		read_pending_state.getParams().setAlgorithmParams(cs);
	}
	public void createMasterSecret(byte[] pre_master_secret) throws Exception {
		write_pending_state.getParams().createMasterSecret(pre_master_secret);
		read_pending_state.getParams().setMasterSecret(write_pending_state.getParams().getMasterSecret());
	}
	public void setMasterSecret(byte[] ms) {
		write_pending_state.getParams().setMasterSecret(ms);
		read_pending_state.getParams().setMasterSecret(ms);
	}

	public int createSecretKeys() throws Exception {
		int len,res,pos=0;
		byte[] buf;
		byte[] rnd = new byte[64];
		TLSStateKeys cl, srv;
		TLSSecurityParameters params = write_pending_state.getParams();
		
		len = (params.enc_key_length + params.mac_key_length + params.fixed_iv_length)*2;
		buf = new byte[len];
		System.arraycopy(params.server_random, 0, rnd, 0, 32);
		System.arraycopy(params.client_random, 0, rnd, 32, 32);
		res = params.prf(params.master_secret, "key expansion", rnd, buf);

		cl = new TLSStateKeys(params.mac_key_length, params.enc_key_length, params.fixed_iv_length);
		srv = new TLSStateKeys(params.mac_key_length, params.enc_key_length, params.fixed_iv_length);

		if(params.mac_key_length != 0) {
			System.arraycopy(buf, pos, cl.macKey, 0, params.mac_key_length);
			pos += params.mac_key_length;
			System.arraycopy(buf, pos, srv.macKey, 0, params.mac_key_length);
			pos += params.mac_key_length;
		}
		if(params.enc_key_length != 0) {
			System.arraycopy(buf, pos, cl.encKey, 0, params.enc_key_length);
			pos += params.enc_key_length;
			System.arraycopy(buf, pos, srv.encKey, 0, params.enc_key_length);
			pos += params.enc_key_length;
		}
		if(params.fixed_iv_length != 0) {
			System.arraycopy(buf, pos, cl.fixedIV, 0, params.fixed_iv_length);
			pos += params.fixed_iv_length;
			System.arraycopy(buf, pos, srv.fixedIV, 0, params.fixed_iv_length);
		}
		
		if(params.isServer == true) {
			write_pending_state.setKeys(srv);
			read_pending_state.setKeys(cl);
		}
		else {
			write_pending_state.setKeys(cl);
			read_pending_state.setKeys(srv);
		}
	                    
		return res;
	}

	public int createSecretKeysCurrentState() throws Exception {
		int len,res,pos=0;
		byte[] buf;
		byte[] rnd = new byte[64];
		TLSStateKeys cl, srv;
		TLSSecurityParameters params = write_current_state.getParams();
		
		len = (params.enc_key_length + params.mac_key_length + params.fixed_iv_length)*2;
		buf = new byte[len];
		System.arraycopy(params.server_random, 0, rnd, 0, 32);
		System.arraycopy(params.client_random, 0, rnd, 32, 32);
		res = params.prf(params.master_secret, "key expansion", rnd, buf);

		cl = new TLSStateKeys(params.mac_key_length, params.enc_key_length, params.fixed_iv_length);
		srv = new TLSStateKeys(params.mac_key_length, params.enc_key_length, params.fixed_iv_length);

		if(params.mac_key_length != 0) {
			System.arraycopy(buf, pos, cl.macKey, 0, params.mac_key_length);
			pos += params.mac_key_length;
			System.arraycopy(buf, pos, srv.macKey, 0, params.mac_key_length);
			pos += params.mac_key_length;
		}
		if(params.enc_key_length != 0) {
			System.arraycopy(buf, pos, cl.encKey, 0, params.enc_key_length);
			pos += params.enc_key_length;
			System.arraycopy(buf, pos, srv.encKey, 0, params.enc_key_length);
			pos += params.enc_key_length;
		}
		if(params.fixed_iv_length != 0) {
			System.arraycopy(buf, pos, cl.fixedIV, 0, params.fixed_iv_length);
			pos += params.fixed_iv_length;
			System.arraycopy(buf, pos, srv.fixedIV, 0, params.fixed_iv_length);
		}
		
		if(params.isServer == true) {
			write_current_state.setKeys(srv);
			read_current_state.setKeys(cl);
		}
		else {
			write_current_state.setKeys(cl);
			read_current_state.setKeys(srv);
		}
	                    
		return res;
	}
}



