package ru.ispras.telecomtesk.tlsv12.TLSState;

import javax.crypto.Mac;
import javax.crypto.spec.SecretKeySpec;

import ru.ispras.telecomtesk.tlsv12.DebugInfo;
import ru.ispras.telecomtesk.tlsv12.TLSAlgorithms.*;
import ru.ispras.telecomtesk.tlsv12.TLSState.TLSState.PaddingFill;

public class TLSSecurityParameters
{
	public boolean isServer = false;
	public TLSCipherSuite cs = TLSCipherSuite.TLS_NULL_WITH_NULL_NULL;
	public KeyExchangeAlgorithm ke_alg = KeyExchangeAlgorithm.NULL;
	public PRFAlgorithm prf = PRFAlgorithm.tls_prf_sha256;
	public BulkCipherAlgorithm bulk_cipher_alg = BulkCipherAlgorithm.bulk_null;
	public CipherType cipher_type = CipherType.stream;
	public byte enc_key_length = 0;
	public byte block_length = 0;
	public byte fixed_iv_length = 0;
	public byte record_iv_length = 0;
	public MACAlgorithm mac_alg = MACAlgorithm.mac_null;
	public byte mac_length = 0;
	public byte mac_key_length = 0;
	public byte compression_alg = 0;
	public byte[] master_secret = new byte[48];
	public byte[] client_random = new byte[32];
	public byte[] server_random = new byte[32];
	
	public TLSSecurityParameters() {}

	/*
	 *      master_secret = PRF(pre_master_secret, "master secret",
     *                    ClientHello.random + ServerHello.random)
     *                    [0..47];
     *      When Diffie-Hellman is used for server authentication and key exchange, 
     *      the negotiated key (Z) is used as the pre_master_secret.
     *      Leading bytes of Z that contain all zero bits are stripped before it is used.
 	 */
	public int createMasterSecret(byte[] pre_master_secret) throws Exception {
		int res = 0, i;
		byte[] rnd = new byte[64];
		byte[] new_pre_master_secret = pre_master_secret;
		boolean isDHalgorithm = KeyExchangeAlgorithm.isDHalgorithm(ke_alg);
		
		if((pre_master_secret != null)&&(pre_master_secret.length != 0)) {
			if(isDHalgorithm == true) {
				for(i=0;i<pre_master_secret.length;i++)
					if(pre_master_secret[i] != 0) break;
				
				if(i != 0) {
					new_pre_master_secret = new byte[pre_master_secret.length - i];
					System.arraycopy(pre_master_secret, i, new_pre_master_secret, 0, pre_master_secret.length - i);
				}
			}
			System.arraycopy(client_random, 0, rnd, 0, 32);
			System.arraycopy(server_random, 0, rnd, 32, 32);
			res = prf(new_pre_master_secret, "master secret", rnd, master_secret);
		}
		else
			res = 10;
		
		return res;
	}
	
	public void setMasterSecret(byte[] ms){
		if(ms != null) {
			master_secret = new byte[ms.length];
			System.arraycopy(ms, 0, master_secret, 0, ms.length);
		}
	}

	public byte[] getMasterSecret(){
		return master_secret;
	}

	public void setRandom(byte[] rnd, boolean isClientRandom){
		if((rnd == null) || (rnd.length != 32))
			System.out.println("TLSSecurityParameters::setRandom: byte[] = NULL or byte[].length != 32");
		else {
			if(isClientRandom == true)
				System.arraycopy(rnd, 0, client_random, 0, 32);
			else
				System.arraycopy(rnd, 0, server_random, 0, 32);
		}
	}

	/*
	 *       P_hash(secret, seed) = HMAC_hash(secret, A(1) + seed) +
     *                        		HMAC_hash(secret, A(2) + seed) +
     *                       		HMAC_hash(secret, A(3) + seed) + ... 
     *                       
     *                       where  A(0) = seed
     *								A(i) = HMAC_hash(secret, A(i-1))
	 */
	public int prf(byte[] secret, String label, byte[] seed, byte[] target) throws Exception {
		int res=0, curpos=0, len, maclen;
		String alg;
		Mac mobj;
		SecretKeySpec sk;
		byte[] curSeed;
		byte[] curHash;
		byte[] lseed;
		
		if((target != null)&&(secret != null)&&(seed != null)) {
			if(target.length == 0) return 0;
			
				switch (prf) {
					case tls_prf_sha256:
						alg = "HmacSHA256";
						break;
					default:
						System.out.println("TLSSecurityParameters::prf: Algorithm not supported");
						return -2;
				}
				
				if(label != null)
					lseed = label.getBytes("ASCII");
				else 
					lseed = new byte[0];
				
				mobj = Mac.getInstance(alg);
				sk = new SecretKeySpec(secret, alg);
				maclen = mobj.getMacLength();
				curSeed = new byte[maclen];
				curHash = new byte[maclen];
				mobj.init(sk);
				// create A(1)
				mobj.update(lseed);
				mobj.update(seed);
				mobj.doFinal(curSeed,0);
				
				// while (curpos != target.length)
				while (true) {
					mobj.update(curSeed);
					mobj.update(lseed);
					mobj.update(seed);
					mobj.doFinal(curHash,0);
					
					len = (maclen <= target.length - curpos) ? maclen : target.length - curpos;
					System.arraycopy(curHash, 0, target, curpos, len);
					curpos += len;
					if(curpos == target.length) break;
					
					// create next A(i)
					mobj.update(curSeed);
					mobj.doFinal(curSeed,0);
				}
		}
		else
			res = -1;
		
		return res;
	}

	// include one byte for 'Padding_length' field
	public int getMinPaddingSize(int data) {
		int res = -1;
		
		if((data >= 0) && (cipher_type == CipherType.block)) {
			if(block_length != 0) 
				res = block_length - data % block_length;
		}
		return res;
	}

	// include one byte for 'Padding_length' field
	public int getMaxPaddingSize(int minPaddingSize) {
		int res = -1;
		
		if(minPaddingSize <= 256)
			res = 256 - ((256 - minPaddingSize) % block_length);
		
		return res;
	}
	
	// include one byte for 'Padding_length' field
	public int getMinPaddingSize(byte[] data) {
		int res = -1;
		
		if((data != null) && (cipher_type == CipherType.block)) {
			if(block_length != 0) 
				res = block_length - data.length % block_length;
		}
		return res;
	}
	
	// include one byte for 'Padding_length' field
	public int getMinPaddingSize(byte[][] data) {
		int res = -1;
		int datalen = 0;
		
		if((data != null) && (cipher_type == CipherType.block)) {
			if(block_length != 0) {
				for(int i=0; i<data.length; i++){
					if(data[i] != null)
						datalen += data[i].length;
				}
				if(datalen != 0)
					res = block_length - datalen % block_length;
			}
		}
		return res;
	}

	public void setAlgorithmParams(TLSCipherSuite cs) {
		this.cs = cs;
		if(cs == TLSCipherSuite.UNKNOWN)
			return;
		
		switch(cs) {
		case TLS_NULL_WITH_NULL_NULL:
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_null; mac_alg = MACAlgorithm.mac_null; break;
		case TLS_RSA_WITH_NULL_MD5:
			ke_alg = KeyExchangeAlgorithm.RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_null; mac_alg = MACAlgorithm.hmac_md5; break;
		case TLS_RSA_WITH_NULL_SHA:
			ke_alg = KeyExchangeAlgorithm.RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_null; mac_alg = MACAlgorithm.hmac_sha1; break;
		case TLS_RSA_WITH_NULL_SHA256:
			ke_alg = KeyExchangeAlgorithm.RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_null; mac_alg = MACAlgorithm.hmac_sha256; break;
		case TLS_RSA_WITH_RC4_128_MD5:
			ke_alg = KeyExchangeAlgorithm.RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_rc4; mac_alg = MACAlgorithm.hmac_md5; break;
		case TLS_RSA_WITH_RC4_128_SHA:
			ke_alg = KeyExchangeAlgorithm.RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_rc4; mac_alg = MACAlgorithm.hmac_sha1; break;
		case TLS_RSA_WITH_3DES_EDE_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_3des; mac_alg = MACAlgorithm.hmac_sha1; break;
		case TLS_RSA_WITH_AES_128_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes128; mac_alg = MACAlgorithm.hmac_sha1; break;
		case TLS_RSA_WITH_AES_256_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes256; mac_alg = MACAlgorithm.hmac_sha1; break;
		case TLS_RSA_WITH_AES_128_CBC_SHA256:
			ke_alg = KeyExchangeAlgorithm.RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes128; mac_alg = MACAlgorithm.hmac_sha256; break;
		case TLS_RSA_WITH_AES_256_CBC_SHA256:
			ke_alg = KeyExchangeAlgorithm.RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes256; mac_alg = MACAlgorithm.hmac_sha256; break;
		
		case TLS_DH_DSS_WITH_3DES_EDE_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.DH_DSS;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_3des; mac_alg = MACAlgorithm.hmac_sha1; break;
		case TLS_DH_RSA_WITH_3DES_EDE_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.DH_RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_3des; mac_alg = MACAlgorithm.hmac_sha1; break;      
		case TLS_DHE_DSS_WITH_3DES_EDE_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.DHE_DSS;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_3des; mac_alg = MACAlgorithm.hmac_sha1; break;     
		case TLS_DHE_RSA_WITH_3DES_EDE_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.DHE_RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_3des; mac_alg = MACAlgorithm.hmac_sha1; break;     
		case TLS_DH_anon_WITH_RC4_128_MD5:
			ke_alg = KeyExchangeAlgorithm.DH_ANON;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_rc4; mac_alg = MACAlgorithm.hmac_md5; break;          
		case TLS_DH_anon_WITH_3DES_EDE_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.DH_ANON;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_3des; mac_alg = MACAlgorithm.hmac_sha1; break;  
		case TLS_DH_DSS_WITH_AES_128_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.DH_DSS;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes128; mac_alg = MACAlgorithm.hmac_sha1; break;       
		case TLS_DH_RSA_WITH_AES_128_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.DH_RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes128; mac_alg = MACAlgorithm.hmac_sha1; break;       
		case TLS_DHE_DSS_WITH_AES_128_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.DHE_DSS;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes128; mac_alg = MACAlgorithm.hmac_sha1; break;      
		case TLS_DHE_RSA_WITH_AES_128_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.DHE_RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes128; mac_alg = MACAlgorithm.hmac_sha1; break;      
		case TLS_DH_anon_WITH_AES_128_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.DH_ANON;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes128; mac_alg = MACAlgorithm.hmac_sha1; break;      
		case TLS_DH_DSS_WITH_AES_256_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.DH_DSS;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes256; mac_alg = MACAlgorithm.hmac_sha1; break;       
		case TLS_DH_RSA_WITH_AES_256_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.DH_RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes256; mac_alg = MACAlgorithm.hmac_sha1; break;       
		case TLS_DHE_DSS_WITH_AES_256_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.DHE_DSS;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes256; mac_alg = MACAlgorithm.hmac_sha1; break;      
		case TLS_DHE_RSA_WITH_AES_256_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.DHE_RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes256; mac_alg = MACAlgorithm.hmac_sha1; break;      
		case TLS_DH_anon_WITH_AES_256_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.DH_ANON;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes256; mac_alg = MACAlgorithm.hmac_sha1; break;   
		case TLS_DH_DSS_WITH_AES_128_CBC_SHA256:
			ke_alg = KeyExchangeAlgorithm.DH_DSS;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes128; mac_alg = MACAlgorithm.hmac_sha256; break; 
		case TLS_DH_RSA_WITH_AES_128_CBC_SHA256:
			ke_alg = KeyExchangeAlgorithm.DH_RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes128; mac_alg = MACAlgorithm.hmac_sha256; break; 
		case TLS_DHE_DSS_WITH_AES_128_CBC_SHA256:
			ke_alg = KeyExchangeAlgorithm.DHE_DSS;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes128; mac_alg = MACAlgorithm.hmac_sha256; break;   
		case TLS_DHE_RSA_WITH_AES_128_CBC_SHA256:
			ke_alg = KeyExchangeAlgorithm.DHE_RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes128; mac_alg = MACAlgorithm.hmac_sha256; break;   
		case TLS_DH_anon_WITH_AES_128_CBC_SHA256:
			ke_alg = KeyExchangeAlgorithm.DH_ANON;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes128; mac_alg = MACAlgorithm.hmac_sha256; break; 
		case TLS_DH_DSS_WITH_AES_256_CBC_SHA256:
			ke_alg = KeyExchangeAlgorithm.DH_DSS;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes256; mac_alg = MACAlgorithm.hmac_sha256; break;    
		case TLS_DH_RSA_WITH_AES_256_CBC_SHA256:
			ke_alg = KeyExchangeAlgorithm.DH_RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes256; mac_alg = MACAlgorithm.hmac_sha256; break;    
		case TLS_DHE_DSS_WITH_AES_256_CBC_SHA256:
			ke_alg = KeyExchangeAlgorithm.DHE_DSS;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes256; mac_alg = MACAlgorithm.hmac_sha256; break;
		case TLS_DHE_RSA_WITH_AES_256_CBC_SHA256:
			ke_alg = KeyExchangeAlgorithm.DHE_RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes256; mac_alg = MACAlgorithm.hmac_sha256; break;
		case TLS_DH_anon_WITH_AES_256_CBC_SHA256:
			ke_alg = KeyExchangeAlgorithm.DH_ANON;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes256; mac_alg = MACAlgorithm.hmac_sha256; break; 
		default:
			System.out.println("TLSSecurityParameters::setAlgorithmParams: TLSCipherSuite not supported");
		}
		
		switch(bulk_cipher_alg) {
		case bulk_null:
			cipher_type = CipherType.stream;
			enc_key_length = 0; block_length = 0; record_iv_length = 0; break;
		case bulk_rc4:
			cipher_type = CipherType.stream;
			enc_key_length = 16; block_length = 0; record_iv_length = 0; break;
		case bulk_3des:
			cipher_type = CipherType.block;
			enc_key_length = 24; block_length = 8; record_iv_length = 8; break;
		case bulk_aes128:
			cipher_type = CipherType.block;
			enc_key_length = 16; block_length = 16; record_iv_length = 16; break;
		case bulk_aes256:
			cipher_type = CipherType.block;
			enc_key_length = 32; block_length = 16; record_iv_length = 16; break;
		default:
			System.out.println("TLSSecurityParameters::setAlgorithmParams: Encrypt algorithm not supported");
		}
		
		switch(mac_alg) {
		case mac_null:
			mac_length = 0; mac_key_length = 0; break;
		case hmac_md5:
			mac_length = 16; mac_key_length = 16; break;
		case hmac_sha1:
			mac_length = 20; mac_key_length = 20; break;
		case hmac_sha256:
			mac_length = 32; mac_key_length = 32; break;
		default:
			System.out.println("TLSSecurityParameters::setAlgorithmParams: MAC algorithm not supported");
		}
	}

	public void printDump(int debugLevel) {
		DebugInfo.printMessage(debugLevel, "TLSSecurityParameters:");
		DebugInfo.printMessage(debugLevel, "   isServer: " + isServer);
		DebugInfo.printMessage(debugLevel, "   TLSCipherSuite: " + cs.name());
		DebugInfo.printMessage(debugLevel, "   KeyExchangeAlgorithm: " + ke_alg.name());
		DebugInfo.printMessage(debugLevel, "   PRFAlgorithm: " + prf.name());
		DebugInfo.printMessage(debugLevel, "   BulkCipherAlgorithm: " + bulk_cipher_alg.name());
		DebugInfo.printMessage(debugLevel, "   CipherType: " + cipher_type.name());
		DebugInfo.printMessage(debugLevel, "   enc_key_length: " + enc_key_length);
		DebugInfo.printMessage(debugLevel, "   block_length: " + block_length);
		DebugInfo.printMessage(debugLevel, "   fixed_iv_length: " + fixed_iv_length);
		DebugInfo.printMessage(debugLevel, "   record_iv_length: " + record_iv_length);
		DebugInfo.printMessage(debugLevel, "   MACAlgorithm: " + mac_alg.name());
		DebugInfo.printMessage(debugLevel, "   mac_length: " + mac_length);
		DebugInfo.printMessage(debugLevel, "   mac_key_length: " + mac_key_length);
		DebugInfo.printMessage(debugLevel, "   compression_alg: " + compression_alg);
		DebugInfo.printDump(debugLevel, master_secret, "   master_secret: ");
		DebugInfo.printDump(debugLevel, client_random, "   client_random: ");
		DebugInfo.printDump(debugLevel, server_random, "   server_random: ");
		DebugInfo.printMessage(debugLevel, "\n");
	}
}
