package ru.ispras.telecomtesk.tlsv12;

import java.util.ArrayList;

import ru.ispras.telecomtesk.tlsv12.DebugInfo;
import ru.ispras.telecomtesk.tlsv12.HandshakeMessages.CertificateVerify;
import ru.ispras.telecomtesk.tlsv12.HandshakeMessages.ClientKeyExchange;
import ru.ispras.telecomtesk.tlsv12.HandshakeMessages.Finished;
import ru.ispras.telecomtesk.tlsv12.HandshakeMessages.HMCertificate;
import ru.ispras.telecomtesk.tlsv12.HandshakeMessages.HandshakeMessage;
import ru.ispras.telecomtesk.tlsv12.HandshakeMessages.Hello;
import ru.ispras.telecomtesk.tlsv12.HandshakeMessages.TLSExtensions.SignatureAndHashAlgorithmExt;
import ru.ispras.telecomtesk.tlsv12.RecordProtocol.ApplicationData;
import ru.ispras.telecomtesk.tlsv12.RecordProtocol.ChangeCipher;
import ru.ispras.telecomtesk.tlsv12.RecordProtocol.RecordLayer;
import ru.ispras.telecomtesk.tlsv12.TLSAlgorithms.HashAlgorithm;
import ru.ispras.telecomtesk.tlsv12.TLSAlgorithms.SignatureAlgorithm;
import ru.ispras.telecomtesk.tlsv12.TLSAlgorithms.SignatureAndHashAlgorithm;
import ru.ispras.telecomtesk.tlsv12.TLSAlgorithms.TLSCipherSuite;
import ru.ispras.telecomtesk.tlsv12.TLSSocket.TLSResponse;
import ru.ispras.telecomtesk.tlsv12.TLSState.TLSConnection;
import ru.ispras.telecomtesk.tlsv12.TLSState.TLSConnectionSelector;
import ru.ispras.telecomtesk.tlsv12.TLSState.TLSServerSpec;

import java.security.KeyStore;
import java.security.PrivateKey;

import ru.ispras.telecomtesk.tlsv12.TlsKeystore;


public class Tlsv12Client {
	protected TLSConnectionSelector conSelector;
	public TLSConnection currentConnection;
	public TLSServerSpec conSpec;

	protected Tlsv12Configure clientConf;
	static String configFileNameDefault = "Tlsv12ClientConfig.txt";

	protected String localKeystore = null;
	protected String localKeystorePwd = null;
	protected KeyStore localks = null;
	protected String CipherSuitesStr = null;
	protected String SAHalgorithmsStr = null;
	protected short[] cipherSuites = null;
	protected SignatureAndHashAlgorithm[] sah = null;

	public Tlsv12Client() {
		this(configFileNameDefault);
	}
	
	public Tlsv12Client(String configFileName) {
		conSelector = new TLSConnectionSelector(null,null,0,0);
		currentConnection = new TLSConnection(conSelector);

		DebugInfo.printMessage(1,"Tlsv12Client: try read config file");
		clientConf = new Tlsv12Configure();
		clientConf.readFromFile(configFileName);

/*
		targetServer = clientConf.getExistingValue("server");
		if (targetServer == null) {
			throw new IllegalArgumentException("Server not specified");
		}
*/

		try {
			CipherSuitesStr = clientConf.getExistingValue("ciphersuits");
		} catch (Exception err) {
			throw new IllegalArgumentException("ciphersuits not specified");
		}

		try {
			SAHalgorithmsStr = clientConf.getExistingValue("algorithms");
		} catch (Exception err) {
			throw new IllegalArgumentException("Signature and Hash algorithms not specified");
		}

		int i;
		String[] csStr;
		String delimeter = ","; // 
		csStr = CipherSuitesStr.split(delimeter);
		cipherSuites = new short[csStr.length];

		try {
			DebugInfo.printMessage(3, "read TLS CipherSuite values:");
			for(i=0; i<cipherSuites.length; i++) {
				cipherSuites[i] = TLSCipherSuite.valueOf(csStr[i]).getValue();
				DebugInfo.printMessage(3, " " + TLSCipherSuite.valueOf(csStr[i]));
			}
		
			String[] sahStr;
			sahStr = SAHalgorithmsStr.split(delimeter);
			sah = new SignatureAndHashAlgorithm[sahStr.length];

			DebugInfo.printMessage(3, "get Signature and Hash algorithms:");
			delimeter = ":"; // 
			String[] sahSubStr;
			for(i=0; i<sah.length; i++) {
				sahSubStr = sahStr[i].split(delimeter);
				sah[i] = new SignatureAndHashAlgorithm(HashAlgorithm.valueOf(sahSubStr[0]),
						SignatureAlgorithm.valueOf(sahSubStr[1]));
				DebugInfo.printMessage(3, " " + sah[i].toString());
			}
		} catch (Exception err) {
			throw new IllegalArgumentException("algorithms illegal value");
		}
	    
		localKeystore = clientConf.getValue("keystore");
		if (localKeystore != null) {
			try {
				localKeystorePwd = clientConf.getExistingValue("keystorePassword");
			} catch (Exception err) {
				throw new IllegalArgumentException("keystorePassword not specified");
			}

			localks = TlsKeystore.getKeyStore(localKeystore, localKeystorePwd);
			if (localks == null) {
				throw new IllegalArgumentException("getting KeyStore failed");
			}
		}
		
		conSpec = new TLSServerSpec();
	}
	
	public byte[] getCurrentSessionID() {
		DebugInfo.printMessage(3, "Tlsv12Client::getCurrentSessionID");
		return currentConnection.getSessionID();
	}

	public void reConnect() {
		DebugInfo.printMessage(3, "Tlsv12Client::reConnect");
		conSpec.disconnect();
		conSpec.connect(currentConnection);
	}

	// create Cliet Hello message
	public byte[] createClientFirstTlsMsg() {
		DebugInfo.printMessage(3, "Tlsv12Client::createFirstTlsMsg");
		int res = 0;
		byte[] resTls = null;

		Hello msg = createClientHello(null,cipherSuites, sah); 

		RecordLayer tlsmsg = new RecordLayer();
		tlsmsg.msglist = new ArrayList();
		tlsmsg.msglist.add(msg);

		conSpec.connect(currentConnection);
		
		res += RecordLayerMsgSpec(tlsmsg);

		if(res == 0)
			resTls = tlsmsg.toByteArray();

		return resTls;
	}

	public byte[] createResumeClientHelloMsg(byte[] id, boolean needReConnect) {
		DebugInfo.printMessage(3, "Tlsv12Client::createResumeClientHelloMsg");
		int res = 0;
		byte[] resTls = null;
		byte[] curSessionID;

		if(id == null)
			curSessionID = currentConnection.getSessionID();
		else
			curSessionID = id;
		
		if(curSessionID == null)
			DebugInfo.printMessage(3, "===Warning!Tlsv12Client::createResumeClientHelloMsg: curSessionID null");
		DebugInfo.printDump(1, curSessionID, "curSessionID");
		
		Hello msg = createClientHello(curSessionID, cipherSuites, sah); 

		RecordLayer tlsmsg = new RecordLayer();
		tlsmsg.msglist = new ArrayList();
		tlsmsg.msglist.add(msg);

		if(needReConnect) reConnect();
		
		res += RecordLayerMsgSpec(tlsmsg);

		if(res == 0)
			resTls = tlsmsg.toByteArray();

		return resTls;
	}

	public byte[] createClientSecondTlsFlow() {
		DebugInfo.printMessage(3, "Tlsv12Client::createSecondTlsFlow");
		if(!conSpec.getCurrentState().connected) {
			DebugInfo.printMessageWithSeparator(1,"===Error! connection state: disconnected");
			return null;
		}
		
		int res = 0;
		byte[] resTls = null;
		ArrayList<RecordLayer> tlsList = new ArrayList<RecordLayer>();
		RecordLayer tlsmsg = new RecordLayer();
		tlsmsg.msglist = new ArrayList();

		if(conSpec.isGetCertificateRequest) {
			DebugInfo.printMessage(1,"Certificate start");
			HMCertificate msgCert = createClientCertificate();
			if(msgCert != null)
				tlsmsg.msglist.add(msgCert);
			else {
				DebugInfo.printMessage(1,"===Error! Certificate return null");
				return null;
			}
		}


		DebugInfo.printMessage(1,"ClientKeyExchange start");
		ClientKeyExchange msgCKE = createClientKeyExchange();
		if(msgCKE != null) {
			tlsmsg.msglist.add(msgCKE);
		} else {
			DebugInfo.printMessage(1,"===Error! ClientKeyExchange return null");
			return null;
		}

		// process messages and add to handshake_messages_list before calculate CertificateVerify signature
		res += RecordLayerMsgSpec(tlsmsg);
		tlsList.add(tlsmsg);

		// 
		if(conSpec.isCertificateVerifyMustSend) {
			DebugInfo.printMessage(1,"CertificateVerify start");
			tlsmsg = new RecordLayer();
			tlsmsg.msglist = new ArrayList();
			CertificateVerify msgCV = createClientCertificateVerify();
			if(msgCV != null) {
				tlsmsg.msglist.add(msgCV);
				res += RecordLayerMsgSpec(tlsmsg);
				tlsList.add(tlsmsg);
			} else {
				DebugInfo.printMessage(1,"===Error! CertificateVerify return null");
				return null;
			}
		}

	
		tlsmsg = new RecordLayer();
		tlsmsg.msglist = new ArrayList();
		
		DebugInfo.printMessage(1,"ChangeCipher start");
		ChangeCipher msgCC = new ChangeCipher();
		tlsmsg.msglist.add(msgCC);

		res += RecordLayerMsgSpec(tlsmsg);
		tlsList.add(tlsmsg);

		
		tlsmsg = new RecordLayer();
		tlsmsg.msglist = new ArrayList();
		
		DebugInfo.printMessage(1,"Finished start");
		Finished msgFn = createClientFinished();
		if(msgFn != null) {
			tlsmsg.msglist.add(msgFn);
		} else res++;

		res += RecordLayerMsgSpec(tlsmsg);
		tlsList.add(tlsmsg);


		
		DebugInfo.printMessageWithSeparator(1,"try combine RecordLayer list");
		if(res == 0) {
			resTls = combineRecordLayerList(tlsList);
		}

		return resTls;
	}

	public Hello createClientHello(byte[] id, short[] cs, SignatureAndHashAlgorithm[] sh) {
		DebugInfo.printMessage(3, "Tlsv12Client::createClientHello");
		if(cs == null)
			DebugInfo.printMessage(3, "cipherSuites null");
		else if(sh == null)
			DebugInfo.printMessage(3, "SignatureAndHashAlgorithms null");
		
		Hello msg = new Hello(id,cs);
		SignatureAndHashAlgorithmExt hl_ext_sah = new SignatureAndHashAlgorithmExt(sh);
		msg.addExtension(hl_ext_sah);

		return msg;
	}
	
	public HMCertificate createClientCertificate() {
		DebugInfo.printMessage(3, "Tlsv12Client::createClientCertificate");
		HMCertificate msgCert = null;
		if(localks != null) {
			DebugInfo.printMessage(3,"local Keystore found");
//			conSpec.localCert = TlsKeystore.getCertificate(localks,target.sah,null); 
//			conSpec.localCert = TlsKeystore.getCertificate(localks,target.sah,SignatureAlgorithm.DSA); 
			conSpec.localCert = TlsKeystore.getCertificate(localks,conSpec.sah,SignatureAlgorithm.RSA);
			if(conSpec.localCert != null) {
				DebugInfo.printMessage(3,"local certificate found");
				conSpec.setClientCertificateType();
				conSpec.isCertificateVerifyMustSend = true;
				msgCert = new HMCertificate(TlsKeystore.getCertificateChainEncoded(localks,conSpec.localCert));
			}
			else {
				DebugInfo.printMessage(3,"local certificate not found");
				msgCert = new HMCertificate(new byte[0][]);
			}
		}
		else {
			DebugInfo.printMessage(3,"local Keystore not found");
			msgCert = new HMCertificate(new byte[0][]);
		}
		return msgCert;
	}

	public ClientKeyExchange createClientKeyExchange() {
		DebugInfo.printMessage(3, "Tlsv12Client::createClientKeyExchange");
		ClientKeyExchange msgCKE = null;
		try {
			msgCKE = new ClientKeyExchange(currentConnection.getWritePendingState().getParams().ke_alg);
			switch (msgCKE.key_alg) {
			case RSA:
				msgCKE.setEncryptedPreMasterSecret(conSpec.serverCert.getPublicKey(),true);
				break;
			default:
				conSpec.pre_master_secret = msgCKE.setDiffieHellmanValue(conSpec.serverDHparams);
			}
		} catch (Exception err) {
			DebugInfo.printMessage(1,"===Error! ClientKeyExchange: " + err);
			return null;
		}

		return msgCKE;
	}
	
	public CertificateVerify createClientCertificateVerify() {
		DebugInfo.printMessage(3, "Tlsv12Client::createClientCertificateVerify");
		CertificateVerify msgCV;
		DebugInfo.printMessage(5,"calcSignature");
		try {
			msgCV = new CertificateVerify(TlsKeystore.getSigAlg(conSpec.localCert));
			PrivateKey localPrivateKey = TlsKeystore.getPrivateKeyCert(localks, conSpec.localCert, localKeystorePwd);
			msgCV.calcSignature( localPrivateKey, 
				conSpec.sessionMsgList, 
				conSpec.last_client_hello_index, 
				conSpec.sessionMsgList.size() - conSpec.last_client_hello_index);
			DebugInfo.printMessage(5,"calcSignature return ");
		}
		catch (Exception err) {
			DebugInfo.printMessage(3,"===Warning! calcSignature failed");
			return null;
		}

		return msgCV;
	}
	

	public Finished createClientFinished() {
		DebugInfo.printMessage(3, "Tlsv12Client::createClientFinished");
		Finished msgFn = new Finished(false);
		try {
			int res = msgFn.calcVerifyData( currentConnection.getWriteCurrentState().getParams(), 
					conSpec.sessionMsgList, 
					conSpec.last_client_hello_index, 
					conSpec.sessionMsgList.size() - conSpec.last_client_hello_index);
			DebugInfo.printMessage(5,"calcVerifyData return " + res);
		}
		catch (Exception err) {
			DebugInfo.printMessage(1,"===Error! calcVerifyData failed");
		}
		return msgFn;
	}

	public ArrayList<RecordLayer> parseTlsResponse(TLSResponse resp) throws Exception {
		return conSpec.receiveTLS(resp, currentConnection);
	}

	public ArrayList<RecordLayer> parseTlsResponse(byte[] tlsmsg) throws Exception {
		TLSResponse resp = new TLSResponse(tlsmsg);
		return this.parseTlsResponse(resp);	
	}

	public byte[] createApplicationData(byte[] data) {
		DebugInfo.printMessage(3, "Tlsv12Client::createApplicationData");
		if(!conSpec.getCurrentState().connected) {
			DebugInfo.printMessageWithSeparator(1,"===Error! connection state: disconnected");
			return null;
		}

		int res = 0;
		byte[] resTls = null;
		ApplicationData msg = new ApplicationData(data); 

		RecordLayer tlsmsg = new RecordLayer();
		tlsmsg.msglist = new ArrayList();
		tlsmsg.msglist.add(msg);

		try {
			if(data.length < 1500)
				tlsmsg.setContent(msg,null,currentConnection);
			else {
				DebugInfo.printMessage(1,"===Warning! Tlsv12Client::createApplicationData: call setLargeContent");
				tlsmsg.setLargeContent(msg,null,currentConnection);
			}
		} catch (Exception err) {
			DebugInfo.printMessage(1,"===Error! Tlsv12Client::createApplicationData:Record layer Exception: " + err);
			res = 1;
		}

		try {
			conSpec.sendApplicationData(tlsmsg,currentConnection);
		} catch (Exception err) {
			DebugInfo.printMessage(1,"===Error! Tlsv12Client::createApplicationData:send Exception: " + err);
		}

		if(res == 0)
			resTls = tlsmsg.toByteArray();

		return resTls;
	}

	public ArrayList<RecordLayer> parseTlsResponseApplicationData(TLSResponse resp) throws Exception {
		return conSpec.receiveApplicationData(resp, currentConnection);
	}

	int RecordLayerMsgSpec(RecordLayer msgRL) {
		int res = 0;
		try {
			DebugInfo.printMessage(3,"Tlsv12Client::RecordLayerMsgSpec:setContent");
			Class chainClass;
			if(msgRL.msglist.get(0) instanceof HandshakeMessage)
				chainClass = HandshakeMessage.class;
			else
				chainClass = msgRL.msglist.get(0).getClass();
			
			DebugInfo.printMessage(3,"Class type: " + chainClass);
			msgRL.setContent(chainClass,msgRL.msglist,0,msgRL.msglist.size(),null,currentConnection);
		} catch (Exception err) {
			DebugInfo.printMessage(1,"===Error! Tlsv12Client::RecordLayerMsgSpec: Record layer Exception: " + err);
			res++;
		}
		msgRL.print();
		try {
			DebugInfo.printMessage(3,"Tlsv12Client::RecordLayerMsgSpec:send");
			conSpec.send(msgRL,currentConnection);
		} catch (Exception err) {
			DebugInfo.printMessage(1,"===Error! Tlsv12Client::RecordLayerMsgSpec: send Exception: " + err);
			res++;
		}
		return res;
	}

	byte[] combineRecordLayerList(ArrayList<RecordLayer> tlsLst) {
		DebugInfo.printMessage(3,"Tlsv12Client::combineRecordLayerList");
		byte[] ba = null;
		int i = 0, fullLength = 0, pos = 0;
		if(tlsLst != null) {
			DebugInfo.printMessage(3,"RecordLayer List size " + tlsLst.size());
			try {
				byte[][] resTlsLst = new byte[tlsLst.size()][];
				for(i=0; i < tlsLst.size(); i++) {
					resTlsLst[i] = tlsLst.get(i).toByteArray();
					if(resTlsLst[i] != null)
						fullLength += resTlsLst[i].length;
				}
				ba = new byte[fullLength];
				for(i=0; i < resTlsLst.length; i++) {
					if(resTlsLst[i] != null) {
						System.arraycopy(resTlsLst[i], 0, ba, pos, resTlsLst[i].length);
						pos += resTlsLst[i].length;
					}
				}
			} catch (Exception err) {
				DebugInfo.printMessage(1,"===Error! Tlsv12Client::combineRecordLayerList: index " + i + ", fullLength " + fullLength + ", position " + pos);
				DebugInfo.printMessage(1,"===Error! Tlsv12Client::combineRecordLayerList: " + err);
			}
		}
		return ba;
	}
}
